###############################################################################
##  Kroller - v0.88
###############################################################################
##  
##  Code by:
##    Brandon Zehm <caspian at dotconf.net>
##  
##  Icon set arranged by:
##    Adam Duskett
##  
##  Kroller is an extension for SuperKaramba to make a MacOS-X like
##  menu bar.
##  It was originally based on tuxbar, but now only a few lines of
##  code from it remain.
##  
##  This SuperKaramba extension was designed to be flexible.
##  You can add new icons, or change existing ones by simply
##  editing the lines in the "User Configurable Section" in this
##  file.  You can use icons of any size you want.  All you need
##  to do is make sure that all the "small" icons are the same
##  size, and all the "big" icons are the same size, and you set
##  those sizes in the imageSize and imageSizeBig variables.
##
##  Feel free to email me with comments or suggestions!
##
###############################################################################

## Import a few standard python modules
import string, re, os, time


## This import statement allows access to the karamba functions
import karamba





############################
##  Global Configuration  ##
############################
conf = {
    
    ## Configuration file location, DO NOT NORMALLY EDIT THIS!  Recent enhancements
    ## in SuperKaramba allow us to locate the conf file on our own.  If you
    ## do hard-code a path here it will still work, but you shouldn't need to.
    "configurationFile"      : "kroller.conf",
    
    "version"                : "v0.88",
    
    ## The following variables are overwritten by the configuration file
    "debug"                  : 0,                         ## Debug level
    "smallImageSize"         : 0,                         ## Size of small images
    "largeImageSize"         : 0,                         ## Size of large images
    "maxFPS"                 : 0,                         ## Maximum widget updates / second.  0 == no limit.
    "animateOnClick"         : 0,                         ## Enable animation when icon is clicked
}


## Define some more global variables
commands         = []
imageFiles       = []
imageFilesBig    = []
images           = []
imagesBig        = []
imageSize        = 0
imageSizeBig     = 0
numIcons         = 0
iconSeparation   = 0
centerSmallIcon  = 0
borderLeft       = 0
borderRight      = 0
borderBottom     = 0
sizeDiff         = 0
sizeRatio        = 0
iconFloor        = 0
raiseIcons       = 0
widgetWidth      = 0
widgetHeight     = 0
sleepTime        = 0
init             = 0
isReset          = 0

resolutionX      = 0
resolutionY      = 0

currentXPosition = 0
currentYPosition = 0
currentSize      = 0

## barimageFile = "pics/osxbarback.png"
## barimage = 0
## barimageBigFile = "pics/osxbarbackBig.png"
## barimageBig = 0













###############################################################################
##  Function:
##    initWidget(widget)
##  
##  Description:
##    This function is buggy.. I think it gets called too early, because
##    if you try to create an image it crashes karamba.
###############################################################################
def initWidget(widget):
    """Initialize Widget"""    
    
    if (conf['debug'] >= 2):
        print "initWidget() got called"
    










###############################################################################
##  Function: 
##    widgetClicked(widget, x, y, button)
##  
##  This gets called everytime our widget is clicked.
##  Notes:
##    widget = reference to our widget
##    x = x position (relative to our widget)
##    y = y position (relative to our widget)
##    botton = button clicked:
##                    1 = Left Mouse Button
##                    2 = Middle Mouse Button
##                    3 = Right Mouse Button, but this will never happen
##                        because the right mouse button brings up the
##                        Karamba menu.
###############################################################################
def widgetClicked(widget, x, y, button):
    """widgetClicked gets called when our widget is clicked"""
    global currentXPosition
    global currentYPosition
    global currentSize
    
    if (conf['debug'] >= 2):
        print "widgetClicked() got called"
    
    
    ## Animate icon if application launched
    if (button == 1 and (int(conf['animateOnClick']) == 1)):
        iconNum = (x - borderLeft) / iconSeparation
        if ((iconNum >= 0) and (iconNum < numIcons)):
            ## Going down
            for down in range (1,raiseIcons,4):
                karamba.moveImage(widget, imagesBig[iconNum], currentXPosition, currentYPosition + down)
                karamba.redrawWidget(widget)
                ## Sleep a bit to accomadate the maxFPS setting
                if (sleepTime):
                    time.sleep(sleepTime)
            ## Going up
            for up in range (1,raiseIcons,4):
                karamba.moveImage(widget, imagesBig[iconNum], currentXPosition, currentYPosition + raiseIcons - up)
                karamba.redrawWidget(widget)
                ## Sleep a bit to accomadate the maxFPS setting
                if (sleepTime):
                    time.sleep(sleepTime)
        
    
    ## Show off if middle clicked (this will be replaced with something else in the future!)
    if (button == 2):
        for i in range (x, (borderLeft + ((numIcons+1) * iconSeparation) - 1), 10):
            widgetMouseMoved(widget, i, (widgetHeight/2), 0)
        for i in range ((borderLeft + ((numIcons+1) * iconSeparation)), (borderLeft-iconSeparation), -10):
            widgetMouseMoved(widget, i, (widgetHeight/2), 0)
        for i in range ((borderLeft-iconSeparation), x, 10):
            widgetMouseMoved(widget, i, (widgetHeight/2), 0)










###############################################################################
##  Function:
##    widgetUpdated(widget)
##  
##  Description:
##    Right now this function is only used for initializing the widget
##
###############################################################################
def widgetUpdated(widget):
    """widgetUpdated gets called at program init and every 4 seconds or so"""    
    global init
    global images
    global imagesBig
    global commands
    global imageFiles
    global imageFilesBig
    global images
    global imagesBig
    global imageSize
    global imageSizeBig
    global numIcons
    global iconSeparation
    global centerSmallIcon
    global borderLeft
    global borderRight
    global borderBottom
    global sizeDiff
    global sizeRatio
    global iconFloor
    global raiseIcons
    global widgetHeight
    global widgetWidth
    global sleepTime
    global resolutionX
    global resolutionY
    
    if (conf['debug'] >= 3):
        print "widgetUpdate got called"
    
    ## Do initialization stuff if we need to
    if init == 0:
        
        ## Read the configuration file
        ## Try hard to find the configuration file, otherwise exit with an error message
        themePath = karamba.getThemePath(widget)
        if (os.path.exists(os.getenv("HOME") + "/.superkaramba/" + conf["configurationFile"])):
            print "Loading configuration from home directory: ", os.getenv("HOME") + "/.superkaramba/" + conf["configurationFile"]
            readConfigurationFile(os.getenv("HOME") + "/.superkaramba/" + conf["configurationFile"])
        elif (os.path.exists(karamba.getThemePath(widget) + "/" + conf["configurationFile"])):
            print "Loading configuration from theme directory: ", karamba.getThemePath(widget) + "/" + conf["configurationFile"]
            readConfigurationFile(karamba.getThemePath(widget) + "/" + conf["configurationFile"])
        elif (os.path.exists(conf["configurationFile"])):
            print "Loading configuration from manually specified location: ", conf["configurationFile"]
            readConfigurationFile(conf["configurationFile"])
        else:
            print ""
            print "ERROR loading the configuration file!  The configuration file could not be located!"
            print "Hint, try putting kroller.conf in ~/.superkaramba/"
            print ""
            sys.exit(1)
        
        ## <calculations>
        
        ## Sizes of your icons (from config file)
        imageSize    = int(conf['smallImageSize'])
        imageSizeBig = int(conf['largeImageSize'])
        
        conf['debug'] = int(conf['debug'])
        
        ## Calculate some other important stuff
        numIcons = len(imageFiles)
        iconSeparation = (imageSize + imageSizeBig)/2            ## The size of the "cells" where icons go
        if (iconSeparation > (2*imageSize)):
            iconSeparation = (2*imageSize)
        centerSmallIcon = ((iconSeparation/2)-(imageSize/2))     ## The number of pixels to center each small icon in it's "cell"
        borderLeft = iconSeparation                              ## How much "white" space on the left
        borderRight = iconSeparation                             ## How much "white" space on the left
        borderBottom = 10                                        ## How much "white" space on the bottom
        sizeDiff = imageSizeBig - imageSize                      ## How many pixels are there from the small icon to the big icon
        sizeRatio = ((sizeDiff + 0.0) / imageSizeBig)            ## Ratio of small icon to big icon
        raiseIcons = 30                                          ## How far to "raise" icons up when icon is at its largest size
        
        ## Calculate widget size
        widgetWidth = (borderLeft + (iconSeparation * numIcons) + borderRight)
        widgetHeight = imageSizeBig + raiseIcons + borderBottom
        
        ## Figure out where the bottom of the icons should be (the iconFloor)
        iconFloor = (widgetHeight - imageSize - borderBottom)
        
        ## Sleep time based on maxFPS from configuration file
        if (conf['maxFPS'] != 0):
            sleepTime = (1.0 / conf['maxFPS'])
        
        if (conf['debug'] >= 1):
            print "numIcons: ", numIcons
            print "imageSize: ", imageSize
            print "imageSizeBig:", imageSizeBig
            print "centerSmallIcon: ", centerSmallIcon
            print "iconSeparation: ", iconSeparation
            print "iconFloor: ", iconFloor
            print "sizeDiff: ", sizeDiff
            print "sizeRatio: ", sizeRatio
            print "borderLeft: ", borderLeft
            print "borderRight: ", borderRight
            print "borderBottom: ", borderBottom
            print "raiseIcons: ", raiseIcons
            print "widgetWidth: ", widgetWidth
            print "widgetHeight: ", widgetHeight
            print "sleepTime: ", sleepTime
            
        ## </calculations>
        
        ## Detect current X resolution
        if (int(conf['detectResolution']) == 1):
            detectResolution()
        else:
            resolutionX = int(conf['resolutionX'])
            resolutionY = int(conf['resolutionY'])
            print "Using specified resolution: " + str(resolutionX) + "x" + str(resolutionY)
        
        ## Set the size of our widget since we're just starting
        karamba.resizeWidget(widget, widgetWidth, widgetHeight)
        
        ## Setup a background image
        ## barimage = karamba.createImage(widget, 60, 62, barimageFile)
        ## barimageBig = karamba.createImage(widget, 6, widgetHeight, barimageBigFile)
        
        for i in range (0,numIcons):
            ## Add small image
            images.append(karamba.createImage(widget, borderLeft + centerSmallIcon + (i*iconSeparation), iconFloor, imageFiles[i]))
            ## Add large image
            imagesBig.append(karamba.createImage(widget, borderLeft + (i*iconSeparation), widgetHeight, imageFilesBig[i]))
            ## Add clickable area that runs a command
            karamba.createClickArea(widget, borderLeft + (i*iconSeparation), (widgetHeight - borderBottom - imageSizeBig - raiseIcons), iconSeparation, imageSizeBig + (raiseIcons/2), commands[i])
            ## Resize the image just in case the image wasn't imageSize pixels wide
            karamba.resizeImage(widget, images[i], imageSize, imageSize)
        
        ## Place widget at the bottom center of the screen
        if (int(conf['autoPosition']) == 1):
            karamba.moveWidget(widget, ((resolutionX/2) - (widgetWidth/2)), (resolutionY-widgetHeight) )
        
        karamba.redrawWidget(widget)
        init = 1
    
#    ## If the cursor isn't currently over our widget, reset it to the default configuration
#    elif (x < (borderLeft-iconSeparation)) or (x >= (borderLeft+(iconSeparation*(numIcons+1)))) or ((y <= 0) and (y >= (widgetHeight - borderBottom))):
#        resetWidget(widget)












###############################################################################
##  Function:   
##    widgetMouseMoved(widget, x, y, button)
##  
##  This gets called everytime the mouse is moved over our widget
##  Notes
##    widget = reference to our widget
##    x = x position (relative to our widget)
##    y = y position (relative to our widget)
##    botton = button being held:
##                    0 = No Mouse Button
##                    1 = Left Mouse Button
##                    2 = Middle Mouse Button
##                    3 = Right Mouse Button, but this will never happen
##                        because the right mouse button brings up the
##                        Karamba menu.
###############################################################################
def widgetMouseMoved(widget, x, y, button):
    """widgetMouseMoved gets called everytime the mouse gets moved over our widget"""    
    global init
    global currentXPosition
    global currentYPosition
    global currentSize
    
    if init == 1:
        iconNum = -2
        
        ## If the mouse is in the target area
        if (x >= (borderLeft-iconSeparation)) and (x < (borderLeft+(iconSeparation*(numIcons+1)))) and ((y >= 0) and (y <= (widgetHeight - borderBottom))):
            
            if (conf['debug'] >= 2):
                print "processing.."
            
            ## Which icon is the mouse over?
            iconNum = (x - borderLeft) / iconSeparation   ## borderLeft is initial white-space, and iconSeparation pixels between each icon
            
            ## This returns a number between 0 and iconSeparation depending on how far through the image you are
            offset              = (x - borderLeft) % iconSeparation         ## 0 - iconSeparation
            offsetRatio         = (offset/(iconSeparation + 0.0) )          ## Ratio from 0-1 of the current offset
            offsetImageSizeBig  = (offsetRatio * (imageSizeBig/2))          ## 0 - (imageSizeBig/2)
            offsetSizeDiff      = (offsetRatio * sizeDiff)                  ## 0 - sizeDiff
            
            ## Unless the current icon comes up as a non-valid number
            if (-2 < iconNum < numIcons+1):
                
                
                ## karamba.moveImage(widget, barimage, 60, 160)                
                ## karamba.moveImage(widget, barimageBig, 6, 62)                
                
                
                ################################
                ## Size and position iconNum-1
                ################################
                
                ## Unless we're on the first image, resize the icon to our left
                if iconNum > 0:
                    ## Calculate stuff
                    ## Size is getting smaller
                    size = imageSize + (sizeDiff/2) - (offsetSizeDiff/2)
                    ## Icon is going left
                    ## If we're positioning the last icon, do a special formula
                    if (iconNum == numIcons):
                        xpos = ( ((iconNum-2) * iconSeparation) + borderLeft + ((iconSeparation/2)-(size/2)) + (iconSeparation/2) + ((iconSeparation/2) * offsetRatio) )
                    else:
                        ##        Initial placement                            center it in it's cell          start half way accross and come back to initial placement
                        xpos = ( ((iconNum-2) * iconSeparation) + borderLeft + ((iconSeparation/2)-(size/2)) + (iconSeparation/2) - ((iconSeparation/2) * offsetRatio) )
                    ## Icon is going down
                    ##     center icon                 start from half way up   and go half way down
                    ypos = (iconFloor + imageSize - size) - (raiseIcons/2) + ((raiseIcons * offsetRatio)/2)
                    
                    ## Place stuff
                    karamba.resizeImage(widget, imagesBig[iconNum-1], size, size)
                    karamba.moveImage(widget, imagesBig[iconNum-1], xpos, ypos)
                    
                    ## Make sure the current small icon is off the screen
                    karamba.moveImage(widget, images[iconNum-1], ((iconNum-1)*iconSeparation)+imageSizeBig, widgetHeight)
                    
                    ## Make sure all the previous icons are setup normally
                    
                    ## If we're positioning the last icon, do a special formula
                    if (iconNum == numIcons):
                        move = offset
                    else:
                        move = 0
                    
                    for i in range (0, iconNum - 1):
                        karamba.moveImage(widget, images[i], ((i-1)*(iconSeparation)) + borderLeft + centerSmallIcon + move, iconFloor)
                        karamba.moveImage(widget, imagesBig[i], ((i-1)*(iconSeparation)) + borderLeft, widgetHeight)
                    
                    if (conf['debug'] >= 2):
                        print "iconNum: ", iconNum-1, "\toffset: ", offset, "\toffsetSizeDiff: ", offsetSizeDiff, "\toffsetImageSizeBig: ", offsetImageSizeBig, "\tsize: ", size, "\txpos: ", xpos, "\typos: ", ypos
                
                ################################
                ## Size and position iconNum
                ################################
                
                if ( (iconNum > -1) and (iconNum < numIcons) ):
                    ## If we're on the left side of the icon
                    if offset < (iconSeparation/2):
                        ## Calculate stuff
                        ## Size is getting larger
                        size = imageSize + (sizeDiff/2) + offsetSizeDiff
                        ## Icon is moving right
                        xpos = ( (iconNum * iconSeparation) + borderLeft + ((iconSeparation/2)-(size/2)) + (iconSeparation/2) - offset )
                        ## Icon is going up
                        ypos = (iconFloor + imageSize - size) - (raiseIcons/2) - ((raiseIcons * offsetRatio)/2)
                        
                        ## Place stuff
                        karamba.resizeImage(widget, imagesBig[iconNum], size, size)
                        karamba.moveImage(widget, imagesBig[iconNum], xpos, ypos)
                    
                    
                    ## If we're on the right side of the icon
                    else:
                        ## Calculate stuff
                        ## Size is getting smaller
                        size = imageSize + sizeDiff + (sizeDiff/2) - offsetSizeDiff
                        ##        initial placement                         center it                       go half way accross and come back offset pixels
                        xpos = ( (iconNum * iconSeparation) + borderLeft + ((iconSeparation/2)-(size/2)) + (iconSeparation/2) - offset )
                        ## icon is going down
                        ypos = (iconFloor + imageSize - size) - raiseIcons + ((raiseIcons * offsetRatio)/2)
                        
                        ## Place stuff
                        karamba.resizeImage(widget, imagesBig[iconNum], size, size)
                        karamba.moveImage(widget, imagesBig[iconNum], xpos, ypos)
                    
                    ## Move the current small image off the widget
                    karamba.moveImage(widget, images[iconNum], (iconNum*iconSeparation) + borderLeft + centerSmallIcon, widgetHeight)
                    
                    ## These are used later if the icon is clicked
                    currentXPosition = xpos
                    currentYPosition = ypos
                    currentSize = size
                    
                    if (conf['debug'] >= 2):
                        print "iconNum: ", iconNum, "\toffset: ", offset, "\toffsetSizeDiff: ", offsetSizeDiff, "\toffsetImageSizeBig: ", offsetImageSizeBig, "\tsize: ", size, "\txpos: ", xpos, "\typos: ", ypos
                    
                
                
                
                
                ################################
                ## Size and position iconNum+1
                ################################
                
                ## Unless we're on the last image, resize the icon to our right
                if (iconNum < numIcons-1):
                    ## Calculate stuff
                    size = imageSize + (offsetSizeDiff/2)
                    ## If we're positioning the first icon, do a special formula
                    if (iconNum == -1):
                        xpos = ( ((iconNum+1) * iconSeparation) + borderLeft + ((iconSeparation/2)-(size/2)) + ((iconSeparation/2) * offsetRatio) )
                    else:
                        ## icon is going left
                        xpos = ( ((iconNum+2) * iconSeparation) + borderLeft + ((iconSeparation/2)-(size/2)) - ((iconSeparation/2) * offsetRatio) )   ## Moves to the left as offset increases
                    ## icon is going up
                    ypos = (iconFloor + imageSize - size) - ((raiseIcons * offsetRatio)/2)
                    
                    
                    ## Place stuff
                    karamba.resizeImage(widget, imagesBig[iconNum+1], size, size)
                    karamba.moveImage(widget, imagesBig[iconNum+1], xpos, ypos)
                    
                    ## Make sure the current small icon is off the screen
                    karamba.moveImage(widget, images[iconNum+1], ((iconNum+1)*iconSeparation)+imageSizeBig, widgetHeight)
                    
                    ## Make sure all the following icons are setup normally
                    
                    ## If we're positioning the first icon, do a special formula
                    if (iconNum == -1):
                        move = offset - iconSeparation
                    else:
                        move = 0
                    
                    for i in range (iconNum+2, numIcons):
                        karamba.moveImage(widget, images[i], ((i)*iconSeparation) + iconSeparation + borderLeft + centerSmallIcon + move, iconFloor)
                        karamba.moveImage(widget, imagesBig[i], ((i)*iconSeparation) + iconSeparation + borderLeft, widgetHeight)
                        
                    if (conf['debug'] >= 2):
                        print "iconNum: ", iconNum+1, "\toffset: ", offset, "\toffsetSizeDiff: ", offsetSizeDiff, "\toffsetImageSizeBig: ", offsetImageSizeBig, "\tsize: ", size, "\txpos: ", xpos, "\typos: ", ypos
                if (conf['debug'] >= 2):
                    print
                
                
                ## Update the screen with the changes we've just made
                karamba.redrawWidget(widget)
                ## Sleep a bit to accomadate the maxFPS setting
                if (sleepTime):
                    time.sleep(sleepTime)
                
        
        ## If mouse is to far right or left reset everything to normal
        else:
            if (conf['debug'] >= 1):
                print "cursor is out of bounds. x: ", x, " y: ", y
            resetWidget(widget)
        










#############################################################################
##  Function:
##    resetWidget(widget)
##  
##  Description:
##    Resets the widget to it's normal state, i.e. no icons zoomed
##  
#############################################################################
def resetWidget(widget):
    """Resets widget to standby state.. i.e. no icons zoomed."""    
    
    for i in range (0,numIcons):
        karamba.moveImage(widget, images[i], (i*iconSeparation) + borderLeft + centerSmallIcon, iconFloor)
        karamba.moveImage(widget, imagesBig[i], (i*iconSeparation) + borderLeft, widgetHeight)
        ## karamba.moveImage(widget, barimage, 60, 62)                
        ## karamba.moveImage(widget, barimageBig, 6, widgetHeight)                
    karamba.redrawWidget(widget)










###############################################################################
## Function:    
##   readConfigurationFile ( string $filename )
## 
## 
## Description: 
##   Reads the file $filename, and stores information from the file into 
##   the global hash %conf.  This version of the function has been modified 
##   specifically to accomadate the needs of Kroller.
## 
## Example: 
##   readConfigurationFile("/etc/configuration.conf");
##
###############################################################################
def readConfigurationFile(filename):
    """Read a configuration file and store name:value pairs into the global variable conf"""
    global commands
    global imageFiles
    global imageFilesBig
    
    ## Open the configuration file
    try:
        file = open(filename,"r")
    except IOError:
        print
        print "ERROR While opening the configuration file [" + filename + "]"
        print "Hint: try putting kroller.conf in ~/.superkaramba/"
        print
    
    ## Read through each line of the file
    while 1:
        line = file.readline()                        ## Read a line from the file
        if len(line) == 0:                            ## Break this while loop if we're at the end of the file
            break
        
        line = re.sub("#.*$", "", line)               ## Remove comments
        line = re.sub("//.*$", "", line)              ## Remove comments
        line = line.strip()                           ## Remove line ending characters, and extra whitespace
        if re.search("\s+=>\s+", line):               ## If it looks like a valid configuration line
            
            ## Debugging
            if (conf['debug'] >= 6):
                print "Line is: ", line
            
            ## Extract the name/value pair
            (name, value) = re.split("\s+=>\s+", line, 1)
            
            ## Add a command
            if (name == "addCommand"):
                commands.append(value)
            
            ## Add a small image
            elif (name == "addSmallImage"):
                imageFiles.append(value)
            
            ## Add a large image
            elif (name == "addLargeImage"):
                imageFilesBig.append(value)
            
            ## Must be a generic configuration parameter, store it in conf
            else:
                if (re.match("^\d+$", value)):
                    conf[name] = int(value)
                else:
                    conf[name] = value
        
    ## Close the coniguration file
    file.close()
    
    
    ## DEBUGGING:
    if conf['debug'] >= 3:
        print "Configuration: "
        for key in conf.keys():
            print key, "\t=>\t", conf[key]










###############################################################################
## Function:    
##   detectResolution ( )
## 
## 
## Description: 
##   Runs xwininfo and sets the current XFree86 resolution to the golbal
##   variables resolutionX and resolutionY
## 
##  Thanks to Adam Geitgey for this code!
###############################################################################
def detectResolution():
    """Detect Current XFree86 Resolution"""
    global resolutionX
    global resolutionY
    
    havexwininfo = os.system("which xwininfo >/dev/null")
    
    if (havexwininfo == 0):
        pass
    else:
        resolutionX = 1024
        resolutionY = 768
        print "\nCan't find xwininfo in your path!  Guessing resolution is 1024x768."
    
    fp = os.popen("xwininfo -root -stats")
    output = fp.read()
    output = output.splitlines()
    i = 0
    for x in output:
        param = x.split()
        if (len(param) > 1):
            if param[0].find("Width:") != -1:
                resolutionX = int(param[1])
            if param[0].find("Height:") != -1:
                resolutionY = int(param[1])
    
    print "Dected resolution: " + str(resolutionX) + "x" + str(resolutionY)









# This will be printed when the widget loads.
print "Loaded my python extension!"

